// I Liceum Oglnoksztacce w Tarnowie
// Koo informatyczne 2007/2011
// (C) mgr Jerzy Waaszek
// Biblioteka procedur graficznych dla trybw 32 bitowych
// Wersja 1.00
//-------------------------------------------------------

#include "newgfx.h"
#include <stack>
#include <list>
#include <cmath>

using namespace std;

//------------------
// DEFINICJE FUNKCJI
//------------------

// Zwraca tekst z numerem wersji biblioteki
//-----------------------------------------

const char * gfxLibVersion(void)
{
  return "1.01\0";
}

// Rysuje piksel 32 bitowy
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne piksela
// color  - 32 bitowy kod koloru piksela
//------------------------------------------------------------------

void gfxPlot(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color)
{
  * ((Uint32 *)screen->pixels + y * screen->w + x) = color;
}

// Rysuje piksel 32 bitowy z przeroczystoci
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne piksela
// color  - 32 bitowy kod koloru piksela
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxPlot(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 alpha)
{
  Uint32 * p = ((Uint32 *)screen->pixels + y * screen->w + x);
  Uint32 c = * p;
  * p = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
        (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
        ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);
}

// Rysuje lini poziom
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne lewego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
//------------------------------------------------------------------

void gfxHLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len)
{
  Uint32 * p = (Uint32 *)screen->pixels + y * screen->w + x;
  for(Uint32 i = 0; i < len; i++) * p++ = color;
}

// Rysuje lini poziom z przeroczystoci
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne lewego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxHLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len, Uint32 alpha)
{
  Uint32 * p = (Uint32 *)screen->pixels + y * screen->w + x;
  Uint32 c;
  for(Uint32 i = 0; i < len; i++)
  {
    c = * p;
    * p ++ = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
             (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
             ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);
  }
}

// Rysuje lini pionow
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne grnego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
//------------------------------------------------------------------

void gfxVLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len)
{
  Uint32 ylen = screen->w;
  Uint32 * p  = (Uint32 *)screen->pixels + y * ylen + x;
  for(Uint32 i = 0; i < len; i++)
  {
    * p = color; p += ylen;
  }
}

// Rysuje lini pionow z przeroczystoci
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne grnego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxVLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len, Uint32 alpha)
{
  Uint32 ylen = screen->w;
  Uint32 * p  = (Uint32 *)screen->pixels + y * ylen + x;
  Uint32 c;
  for(Uint32 i = 0; i < len; i++)
  {
    c   = * p;
    * p = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
           (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
           ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);
    p += ylen;
  }
}

// Rysuje odcinek pomidzy zadanymi punktami
// screen - wskanik struktury SDL_Surface
// x1,y1  - wsprzdne punktu startowego
// x2,y2  - wsprzdne punktu kocowego
// color  - kolor odcinka
//------------------------------------------------------------------

void gfxLine(SDL_Surface * screen, Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2, Uint32 color)
{
  Uint8 * p;
  int dx, dy, kx, ky, e;

  kx = (x1 <= x2) ? 4 : -4;
  ky = (y1 <= y2) ? screen->pitch : -screen->pitch;

  dx = x2 - x1; if(dx < 0) dx = -dx;
  dy = y2 - y1; if(dy < 0) dy = -dy;

  p = (Uint8 *)screen->pixels + y1 * screen->pitch + (x1 << 2);
  * (Uint32 *) p = color;

  if(dx >= dy)
  {
    e = dx >> 1;
    for(int i = 0; i < dx; i++)
    {
      p += kx;  e -= dy;
      if(e < 0)
      {
        p += ky; e += dx;
      }
      * (Uint32 *) p = color;
    }
  }
  else
  {
    e = dy >> 1;
    for(int i = 0; i < dy; i++)
    {
      p += ky; e -= dx;
      if(e < 0)
      {
        p += kx; e += dy;
      }
      * (Uint32 *) p = color;
    }
  }
}

// Rysuje odcinek pomidzy zadanymi punktami z przeroczystoci
// screen - wskanik struktury SDL_Surface
// x1,y1  - wsprzdne punktu startowego
// x2,y2  - wsprzdne punktu kocowego
// color  - kolor odcinka
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxLine(SDL_Surface * screen, Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2, Uint32 color, Uint32 alpha)
{
  Uint8 * p;
  int dx, dy, kx, ky, e;
  Uint32 c;

  kx = (x1 <= x2) ? 4 : -4;
  ky = (y1 <= y2) ? screen->pitch : -screen->pitch;

  dx = x2 - x1; if(dx < 0) dx = -dx;
  dy = y2 - y1; if(dy < 0) dy = -dy;

  p = (Uint8 *)screen->pixels + y1 * screen->pitch + (x1 << 2);
  c   = * (Uint32 * ) p;
  * (Uint32 *) p = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
                   (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
                   ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);

  if(dx >= dy)
  {
    e = dx >> 1;
    for(int i = 0; i < dx; i++)
    {
      p += kx;  e -= dy;
      if(e < 0)
      {
        p += ky; e += dx;
      }
      c   = * (Uint32 * ) p;
      * (Uint32 *) p = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
                       (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
                       ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);
    }
  }
  else
  {
    e = dy >> 1;
    for(int i = 0; i < dy; i++)
    {
      p += ky; e -= dx;
      if(e < 0)
      {
        p += kx; e += dy;
      }
      c   = * (Uint32 * ) p;
      * (Uint32 *) p = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
                       (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
                       ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);
    }
  }
}

// wsprzdne pocztku odcinka dla gfxLineTo
//------------------------------------------------------------------

Sint32 gfx_x_coord = 0;
Sint32 gfx_y_coord = 0;

// Ustawia wsprzdne pocztku odcinka
// x,y - wsprzdne dla gfx_x_coord i gfx_y_coord
//------------------------------------------------------------------

void gfxMoveTo(Sint32 x, Sint32 y)
{
     gfx_x_coord = x; gfx_y_coord = y;
}

// Rysuje odcinek od zapamitanych wsprzdnych
// do nowych wsprzdnych, ktre po operacji s zapamitywane.
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne koca odcinka
// color  - kod koloru odcinka
//------------------------------------------------------------------

void gfxLineTo(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color)
{
   gfxLine(screen, gfx_x_coord, gfx_y_coord, x, y, color);
   gfx_x_coord = x; gfx_y_coord = y;
}

// Rysuje odcinek z przeroczystoci od zapamitanych wsprzdnych
// do nowych wsprzdnych, ktre po operacji s zapamitywane.
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne koca odcinka
// color  - kod koloru odcinka
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxLineTo(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 alpha)
{
   Uint32 c = gfxGetPixel(screen,x,y);
   gfxLine(screen, gfx_x_coord, gfx_y_coord, x, y, color, alpha);
   gfxPlot(screen,x,y,c);
   gfx_x_coord = x; gfx_y_coord = y;
}

// Rysuje piksel 32 bitowy wewntrz zdefiniowanego
// prostokta obcinajcego clip_rect
// screen - wskanik struktury SDL_Surface
// x,y - wsprzdne piksela
// color - 32 bitowy kod koloru piksela
//------------------------------------------------------------------

void gfxClipPlot(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color)
{
  if((x >= screen->clip_rect.x) && (x < screen->clip_rect.x + screen->clip_rect.w) &&
     (y >= screen->clip_rect.y) && (y < screen->clip_rect.y + screen->clip_rect.h))
    * ((Uint32 *)screen->pixels + y * screen->w + x) = color;
}

// Rysuje piksel 32 bitowy z przeroczystoci wewntrz zdefiniowanego
// prostokta obcinajcego clip_rect
// screen - wskanik struktury SDL_Surface
// x,y - wsprzdne piksela
// color - 32 bitowy kod koloru piksela
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxClipPlot(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 alpha)
{
  if((x >= screen->clip_rect.x) && (x < screen->clip_rect.x + screen->clip_rect.w) &&
     (y >= screen->clip_rect.y) && (y < screen->clip_rect.y + screen->clip_rect.h))
  {
    Uint32 * p = ((Uint32 *)screen->pixels + y * screen->w + x);
    Uint32 c = * p;
    * p = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
          (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
          ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);
  }
}

// Rysuje lini poziom z obciciem do clip_rect
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne lewego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
//------------------------------------------------------------------

void gfxClipHLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len)
{
  Sint32 x1c, x2c, y1c, y2c, xp;

  x1c = screen->clip_rect.x;
  y1c = screen->clip_rect.y;
  x2c = x1c + screen->clip_rect.w;
  y2c = y1c + screen->clip_rect.h;

// obcinanie odcinka

  xp = x + len;
  if((y >= y1c) && (xp > x1c) && (x < x2c) && (y < y2c))
  {
    if(x < x1c)
    {
      len += x - x1c;
      x = x1c;
    }
    if(xp > x2c) len = x2c - x;
    gfxHLine(screen, x, y, color, len);
  }
}

// Rysuje lini poziom z obciciem do clip_rect z przeroczystoci
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne lewego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxClipHLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len, Uint32 alpha)
{
  Sint32 x1c, x2c, y1c, y2c, xp;

  x1c = screen->clip_rect.x;
  y1c = screen->clip_rect.y;
  x2c = x1c + screen->clip_rect.w;
  y2c = y1c + screen->clip_rect.h;

// obcinanie odcinka

  xp = x + len;
  if((y >= y1c) && (xp > x1c) && (x < x2c) && (y < y2c))
  {
    if(x < x1c)
    {
      len += x - x1c;
      x = x1c;
    }
    if(xp > x2c) len = x2c - x;
    gfxHLine(screen, x, y, color, len, alpha);
  }
}

// Rysuje lini pionow z obciciem do prostokta clip_rect
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne grnego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
//------------------------------------------------------------------

void gfxClipVLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len)
{
  Sint32 x1c, x2c, y1c, y2c, yp;

  x1c = screen->clip_rect.x;
  y1c = screen->clip_rect.y;
  x2c = x1c + screen->clip_rect.w;
  y2c = y1c + screen->clip_rect.h;

// obcinanie odcinka

  yp = y + len;
  if((x >= x1c) && (yp > y1c) && (y < y2c) && (x < x2c))
  {
    if(y < y1c)
    {
      len += y - y1c;
      y = y1c;
    }
    if(yp > y2c) len = y2c - y;
    gfxVLine(screen, x, y, color, len);
  }
}

// Rysuje lini pionow z obciciem do prostokta clip_rect z przeroczystoci
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne grnego pocztku lini
// color  - 32 bitowy kod koloru linii
// len    - dugo linii w pikselach
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxClipVLine(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 len, Uint32 alpha)
{
  Sint32 x1c, x2c, y1c, y2c, yp;

  x1c = screen->clip_rect.x;
  y1c = screen->clip_rect.y;
  x2c = x1c + screen->clip_rect.w;
  y2c = y1c + screen->clip_rect.h;

// obcinanie odcinka

   yp = y + len;
  if((x >= x1c) && (yp > y1c) && (y < y2c) && (x < x2c))
  {
    if(y < y1c)
    {
      len += y - y1c;
      y = y1c;
    }
    if(yp > y2c) len = y2c - y;
    gfxVLine(screen, x, y, color, len, alpha);
  }
}

// Rysuje ramk na obrysie prostokta
// screen - wskanik struktury SDL_Surface
// r      - definiuje prostokt ramki
// color  - 32 bitowy kod koloru ramki
//------------------------------------------------------------------

void gfxRect(SDL_Surface * screen, SDL_Rect * r, Uint32 color)
{
  if(r)
  {
    if(r->w) gfxHLine(screen, r->x, r->y, color, r->w);
    if((r->w) && (r->h > 1)) gfxHLine(screen, r->x, r->y + r->h - 1, color, r->w);
    if(r->h > 2) gfxVLine(screen, r->x, r->y + 1, color, r->h - 2);
    if((r->h > 2) && (r->w > 1)) gfxVLine(screen, r->x + r->w - 1, r->y + 1, color, r->h - 2);
  }
  else
  {
    gfxHLine(screen, 0, 0, color, screen->w);
    gfxHLine(screen, 0, screen->h - 1, color, screen->w);
    gfxVLine(screen, 0, 1, color, screen->h - 2);
    gfxVLine(screen, screen->w - 1, 1, color, screen->h - 2);
  }
}

// Rysuje ramk na obrysie prostokta z przeroczystoci
// screen - wskanik struktury SDL_Surface
// r      - definiuje prostokt ramki
// color  - 32 bitowy kod koloru ramki
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxRect(SDL_Surface * screen, SDL_Rect * r, Uint32 color, Uint32 alpha)
{
  if(r)
  {
    if(r->w)     gfxHLine(screen, r->x, r->y, color, r->w, alpha);
    if((r->w) && (r->h > 1)) gfxHLine(screen, r->x, r->y + r->h - 1, color, r->w, alpha);
    if(r->h > 2) gfxVLine(screen, r->x, r->y + 1, color, r->h - 2, alpha);
    if((r->h > 2) && (r->w > 1)) gfxVLine(screen, r->x + r->w - 1, r->y + 1, color, r->h - 2, alpha);
  }
  else
  {
    gfxHLine(screen, 0, 0, color, screen->w, alpha);
    gfxHLine(screen, 0, screen->h - 1, color, screen->w, alpha);
    gfxVLine(screen, 0, 1, color, screen->h - 2, alpha);
    gfxVLine(screen, screen->w - 1, 1, color, screen->h - 2, alpha);
  }
}

// Rysuje ramk na obrysie prostokta, reaguje na prostokat obcinania
// screen - wskanik struktury SDL_Surface
// r      - definiuje prostokt ramki
// color  - 32 bitowy kod koloru ramki
//------------------------------------------------------------------

void gfxClipRect(SDL_Surface * screen, SDL_Rect * r, Uint32 color)
{
  if(r)
  {
    if(r->w)     gfxClipHLine(screen, r->x, r->y, color, r->w);
    if((r->w) && (r->h > 1)) gfxClipHLine(screen, r->x, r->y + r->h - 1, color, r->w);
    if(r->h > 2) gfxClipVLine(screen, r->x, r->y + 1, color, r->h - 2);
    if((r->h > 2) && (r->w > 1)) gfxClipVLine(screen, r->x + r->w - 1, r->y + 1, color, r->h - 2);
  }
  else
  {
    gfxClipHLine(screen, 0, 0, color, screen->w);
    gfxClipHLine(screen, 0, screen->h - 1, color, screen->w);
    gfxClipVLine(screen, 0, 1, color, screen->h - 2);
    gfxClipVLine(screen, screen->w - 1, 1, color, screen->h - 2);
  }
}

// Rysuje ramk z przeroczystoci na obrysie prostokta z obcinaniem
// screen - wskanik struktury SDL_Surface
// r      - definiuje prostokt ramki
// color  - 32 bitowy kod koloru ramki
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxClipRect(SDL_Surface * screen, SDL_Rect * r, Uint32 color, Uint32 alpha)
{
  if(r)
  {
    if(r->w)     gfxClipHLine(screen, r->x, r->y, color, r->w, alpha);
    if((r->w) && (r->h > 1)) gfxClipHLine(screen, r->x, r->y + r->h - 1, color, r->w, alpha);
    if(r->h > 2) gfxClipVLine(screen, r->x, r->y + 1, color, r->h - 2, alpha);
    if((r->h > 2) && (r->w > 1)) gfxClipVLine(screen, r->x + r->w - 1, r->y + 1, color, r->h - 2, alpha);
  }
  else
  {
    gfxClipHLine(screen, 0, 0, color, screen->w, alpha);
    gfxClipHLine(screen, 0, screen->h - 1, color, screen->w, alpha);
    gfxClipVLine(screen, 0, 1, color, screen->h - 2, alpha);
    gfxClipVLine(screen, screen->w - 1, 1, color, screen->h - 2, alpha);
  }
}

// Rysuje odcinek pomidzy zadanymi punktami obcity do
// obszaru prostokta obcinajcego ustawionego przez SDL_SetClipRect()
// screen - wskanik struktury SDL_Surface
// x1,y1  - wsprzdne punktu startowego
// x2,y2  - wsprzdne punktu kocowego
// color  - kolor odcinka
//------------------------------------------------------------------

void gfxClipLine(SDL_Surface * screen, Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2, Uint32 color)
{
  Sint32 x1c, x2c, y1c, y2c, iax;

  x1c = screen->clip_rect.x; x2c = x1c + screen->clip_rect.w - 1;
  y1c = screen->clip_rect.y; y2c = y1c + screen->clip_rect.h - 1;
  for(;;)
  {
    Uint32 k1, k2, uax;
    k1 = (((x1 - x1c) >> 31) & 1) | (((x2c - x1) >> 30) & 2) |
         (((y1 - y1c) >> 29) & 4) | (((y2c - y1) >> 28) & 8);
    k2 = (((x2 - x1c) >> 31) & 1) | (((x2c - x2) >> 30) & 2) |
         (((y2 - y1c) >> 29) & 4) | (((y2c - y2) >> 28) & 8);
    if(!(k1 | k2))
    {
      gfxLine(screen, x1, y1, x2, y2, color); break;
    }
    if(k1 & k2) break;
    if(!k1)
    {
      uax = k1; k1 = k2; k2 = uax;
      iax = x1; x1 = x2; x2 = iax;
      iax = y1; y1 = y2; y2 = iax;
    }
    if(k1 & 1)
    {
      y1 += ((y2 - y1) * (x1c - x1)) / (x2 - x1);
      x1 = x1c;
    }
    else if(k1 & 2)
    {
      y1 += ((y2 - y1) * (x1 - x2c)) / (x1 - x2);
      x1 = x2c;
    }
    else if(k1 & 4)
    {
      x1 += ((x1 - x2) * (y1 - y1c)) / (y2 - y1);
      y1 = y1c;
    }
    else
    {
      x1 += ((x1 - x2) * (y1 - y2c)) / (y2 - y1);
      y1 = y2c;
    }
  }
}

// Rysuje odcinek z przeroczystoci pomidzy zadanymi punktami, obcity do
// obszaru prostokta obcinajcego ustawionego przez SDL_SetClipRect()
// screen - wskanik struktury SDL_Surface
// x1,y1  - wsprzdne punktu startowego
// x2,y2  - wsprzdne punktu kocowego
// color  - kolor odcinka
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxClipLine(SDL_Surface * screen, Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2, Uint32 color, Uint32 alpha)
{
  Sint32 x1c, x2c, y1c, y2c, iax;

  x1c = screen->clip_rect.x; x2c = x1c + screen->clip_rect.w - 1;
  y1c = screen->clip_rect.y; y2c = y1c + screen->clip_rect.h - 1;
  for(;;)
  {
    Uint32 k1, k2, uax;
    k1 = (((x1 - x1c) >> 31) & 1) | (((x2c - x1) >> 30) & 2) |
         (((y1 - y1c) >> 29) & 4) | (((y2c - y1) >> 28) & 8);
    k2 = (((x2 - x1c) >> 31) & 1) | (((x2c - x2) >> 30) & 2) |
         (((y2 - y1c) >> 29) & 4) | (((y2c - y2) >> 28) & 8);
    if(!(k1 | k2))
    {
      gfxLine(screen, x1, y1, x2, y2, color, alpha); break;
    }
    if(k1 & k2) break;
    if(!k1)
    {
      uax = k1; k1 = k2; k2 = uax;
      iax = x1; x1 = x2; x2 = iax;
      iax = y1; y1 = y2; y2 = iax;
    }
    if(k1 & 1)
    {
      y1 += ((y2 - y1) * (x1c - x1)) / (x2 - x1);
      x1 = x1c;
    }
    else if(k1 & 2)
    {
      y1 += ((y2 - y1) * (x1 - x2c)) / (x1 - x2);
      x1 = x2c;
    }
    else if(k1 & 4)
    {
      x1 += ((x1 - x2) * (y1 - y1c)) / (y2 - y1);
      y1 = y1c;
    }
    else
    {
      x1 += ((x1 - x2) * (y1 - y2c)) / (y2 - y1);
      y1 = y2c;
    }
  }
}

// Rysuje odcinek od zapamitanych wsprzdnych
// do nowych wsprzdnych, ktre po operacji s zapamitywane. Odcinek
// jest obcinany do prostokta obcinania
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne koca odcinka
// color  - kod koloru odcinka
// clip   - wskanik struktury SDL_Rect z prostoktem obcinajcym
//------------------------------------------------------------------

void gfxClipLineTo(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color)
{
   gfxClipLine(screen, gfx_x_coord, gfx_y_coord, x, y, color);
   gfx_x_coord = x; gfx_y_coord = y;
}

// Rrysuje odcinek z przeroczystoci od zapamitanych wsprzdnych
// do nowych wsprzdnych, ktre po operacji s zapamitywane. Odcinek
// jest obcinany do prostokta obcinania. Ostatni punkt, ze wzgldu na przeroczysto,
// nie jest rysowany.
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne koca odcinka
// color  - kod koloru odcinka
// clip   - wskanik struktury SDL_Rect z prostoktem obcinajcym
// alpha  - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//------------------------------------------------------------------

void gfxClipLineTo(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color, Uint32 alpha)
{

   Uint32 c = 0;
   if((x >= screen->clip_rect.x) &&
      (x  < screen->clip_rect.x + screen->clip_rect.w) &&
      (y >= screen->clip_rect.y) &&
      (y <  screen->clip_rect.y + screen->clip_rect.h))
   {
     c = gfxGetPixel(screen,x,y);
     gfxClipLine(screen, gfx_x_coord, gfx_y_coord, x, y, color, alpha);
     gfxClipPlot(screen,x,y,c);
   }
   else gfxClipLine(screen, gfx_x_coord, gfx_y_coord, x, y, color, alpha);

   gfx_x_coord = x; gfx_y_coord = y;
}

// gfxGetPixel() odczytuje 32 bitowy kod piksela
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne piksela
//------------------------------------------------------------------

Uint32 gfxGetPixel(SDL_Surface * screen, Sint32 x, Sint32 y)
{
  return * ((Uint32 *)screen->pixels + y * screen->w + x);
}

// gfxBoundaryFill wypenia obszar ograniczony prostoktem obcinania
// zawartym w strukturze SDL_Surface oraz konturem
// screen - wskanik struktury SDL_Surface
// x,y    - adres ziarna pocztkowego
// fcolor - kolor wypeniania
// bcolor - kolor konturu - brzegu
//------------------------------------------------------------------

void gfxBoundaryFill(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 fcolor, Uint32 bcolor)
{
  Sint32 cx1 = screen->clip_rect.x;
  Sint32 cy1 = screen->clip_rect.y;
  Sint32 cx2 = cx1 + screen->clip_rect.w - 1;
  Sint32 cy2 = cy1 + screen->clip_rect.h - 1;
  Sint32 x1;
  Sint32 x2 = x;
  Sint32 bx1;
  Sint32 bx2;
  Uint32 * p1 = (Uint32 *)screen->pixels + y * screen->w + x;
  Uint32 * p2 = p1;
  Sint32 len = x2 - x;
  stack<Sint32> q;

  while((--x >= cx1) && (*(--p1) != bcolor));

  x++;

  while((++x2 <= cx2) && (*(++p2) != bcolor));

  len = x2 - x;
  q.push(x); q.push(y); q.push(len);

  while(!q.empty())
  {
    len = q.top(); q.pop();
    y   = q.top(); q.pop();
    x   = q.top(); q.pop();

    gfxHLine(screen, x, y, fcolor, len);

    if(y > cy1)
    {
      x1 = x;
      x2 = x1 + len - 1;

      p1 = (Uint32 *)screen->pixels + (y - 1) * screen->w + x1;

      while(x1 <= x2)
      {
        if(* p1 == fcolor)
        {
          x1 += 2; p1 += 2;
        }
        else if(* p1 == bcolor)
        {
          x1++; p1++;
        }
        else
        {
          bx1 = x1;
          bx2 = x1;
          p2 = p1;

          while((--bx1 >= cx1) && (*(--p2) != bcolor));

          bx1++;
          p2 = p1;

          while((++bx2 <= cx2) && (*(++p2) != bcolor));

          q.push(bx1); q.push(y - 1); q.push(bx2 - bx1);
          x1 = bx2 + 1; p1 = p2 + 1;
        }
      }
    }
    if(y < cy2)
    {
      x1 = x;
      x2 = x1 + len - 1;

      p1 = (Uint32 *)screen->pixels + (y + 1) * screen->w + x1;

      while(x1 <= x2)
      {
        if(* p1 == fcolor)
        {
          x1 += 2; p1 += 2;
        }
        else if(* p1 == bcolor)
        {
          x1++; p1++;
        }
        else
        {
          bx1 = x1;
          bx2 = x1;
          p2 = p1;

          while((--bx1 >= cx1) && (*(--p2) != bcolor));

          bx1++;
          p2 = p1;

          while((++bx2 <= cx2) && (*(++p2) != bcolor));

          q.push(bx1); q.push(y + 1); q.push(bx2 - bx1);
          x1 = bx2 + 1; p1 = p2 + 1;
        }
      }
    }
  }
}

// gfxFloodFill wypenia obszar ograniczony prostoktem obcinania
// zawartym w strukturze SDL_Surface oraz obszarami w innych kolorach
// screen - wskanik struktury SDL_Surface
// x,y    - adres ziarna pocztkowego
// fcolor - kolor wypeniania
//------------------------------------------------------------------

void gfxFloodFill(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 fcolor)
{
  Sint32 cx1 = screen->clip_rect.x;
  Sint32 cy1 = screen->clip_rect.y;
  Sint32 cx2 = cx1 + screen->clip_rect.w - 1;
  Sint32 cy2 = cy1 + screen->clip_rect.h - 1;
  Sint32 x1;
  Sint32 x2 = x;
  Sint32 bx1;
  Sint32 bx2;
  Uint32 * p1 = (Uint32 *)screen->pixels + y * screen->w + x;
  Uint32 * p2 = p1;
  Uint32 rcolor = * p1;
  Sint32 len;
  stack<Sint32> q;

  while((--x >= cx1) && (*(--p1) == rcolor));

  x++;

  while((++x2 <= cx2) && (*(++p2) == rcolor));

  len = x2 - x;
  q.push(x); q.push(y); q.push(len);

  while(!q.empty())
  {
    len = q.top(); q.pop();
    y   = q.top(); q.pop();
    x   = q.top(); q.pop();

    gfxHLine(screen, x, y, fcolor, len);

    if(y > cy1)
    {
      x1 = x;
      x2 = x1 + len - 1;

      p1 = (Uint32 *)screen->pixels + (y - 1) * screen->w + x1;

      while(x1 <= x2)
      {
        if(* p1 == fcolor)
        {
          x1 += 2; p1 += 2;
        }
        else if(* p1 != rcolor)
        {
          x1++; p1++;
        }
        else
        {
          bx1 = x1;
          bx2 = x1;
          p2 = p1;

          while((--bx1 >= cx1) && (*(--p2) == rcolor));

          bx1++;
          p2 = p1;

          while((++bx2 <= cx2) && (*(++p2) == rcolor));

          q.push(bx1); q.push(y - 1); q.push(bx2 - bx1);
          x1 = bx2 + 1; p1 = p2 + 1;
        }
      }
    }
    if(y < cy2)
    {
      x1 = x;
      x2 = x1 + len - 1;

      p1 = (Uint32 *)screen->pixels + (y + 1) * screen->w + x1;

      while(x1 <= x2)
      {
        if(* p1 == fcolor)
        {
          x1 += 2; p1 += 2;
        }
        else if(* p1 != rcolor)
        {
          x1++; p1++;
        }
        else
        {
          bx1 = x1;
          bx2 = x1;
          p2 = p1;

          while((--bx1 >= cx1) && (*(--p2) == rcolor));

          bx1++;
          p2 = p1;

          while((++bx2 <= cx2) && (*(++p2) == rcolor));

          q.push(bx1); q.push(y + 1); q.push(bx2 - bx1);
          x1 = bx2 + 1; p1 = p2 + 1;
        }
      }
    }
  }
}

// Wypenia zadany wielokt podanym kolorem, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
//-----------------------------------------

// Element list LK i LKA
//----------------------

typedef struct
{
  Sint32 y;
  double x, dx;
} LKElement;

// Funkcja suy do porwna elementw list LK i LKA
//--------------------------------------------------

bool CompareLKElements(LKElement a, LKElement b)
{
   return a.x < b.x;
}

void gfxFillPoly(SDL_Surface * s, Sint32 * f, Uint32 color)
{
  list<LKElement> LK[s->clip_rect.h];  // tablica kubekw
  list<LKElement> LKA;                 // lista krawdzi aktywnych

  Sint32 y1clip = s->clip_rect.y;
  Sint32 y2clip = y1clip + s->clip_rect.h - 1;
  Sint32 ymax   = y1clip;
  Sint32 ymin   = y2clip;
  Sint32 xp, xk, yp, yk, x1, x2, y1, y2, y, plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    while(--plen)
    {
      xk = * f++; yk = * f++;
      if(yp > yk)
      {
        x1 = xk; y1 = yk; x2 = xp; y2 = yp;
      }
      else
      {
        x1 = xp; y1 = yp; x2 = xk; y2 = yk;
      }
      if(((y1 >= y1clip) || (y2 >= y1clip)) && ((y1 <= y2clip) || (y2 <= y2clip)))
      {
        LKElement E;
        if(ymin > y1) ymin = y1;
        if(ymax < y2) ymax = y2;
        E.x  = x1;
        E.dx = (x2 - x1) / (double)(y2 - y1);
        if(y1 < y1clip)
        {
          E.x += E.dx * (y1clip - y1);
          y1 = y1clip;
        }
        E.y = y2;
        if(y1 != y2) LK[y1 - y1clip].push_front(E);
      }
      xp = xk; yp = yk;
    }
  }

  if(ymin < y1clip) ymin = y1clip;
  if(ymax > y2clip) ymax = y2clip;

  for(y = ymin; y <= ymax; y++)
  {
    list<LKElement>::iterator i = LKA.begin();
    while(i != LKA.end()) if(i->y == y) i = LKA.erase(i); else i++;
    if(LK[y - y1clip].size())
    {
      LK[y - y1clip].sort(CompareLKElements);
      LKA.merge(LK[y - y1clip],CompareLKElements);
    }

    bool selector = false;

    for(list<LKElement>::iterator i = LKA.begin(); i != LKA.end(); i++)
    {
      if(selector)
      {
        x2 = (Sint32)floor(i->x);
        gfxClipHLine(s, x1, y, color, x2 - x1 + 1);
      }
      else x1 = (Sint32)ceil(i->x);
      selector = !selector;
      i->x += i->dx;
    }
    LKA.sort(CompareLKElements);
  }
  LKA.clear();
}

// Wypenia zadany wielokt podanym kolorem z przeroczystoci, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//-----------------------------------------

void gfxFillPoly(SDL_Surface * s, Sint32 * f, Uint32 color, Uint32 alpha)
{
  list<LKElement> LK[s->clip_rect.h];  // tablica kubekw
  list<LKElement> LKA;                 // lista krawdzi aktywnych

  Sint32 y1clip = s->clip_rect.y;
  Sint32 y2clip = y1clip + s->clip_rect.h - 1;
  Sint32 ymax   = y1clip;
  Sint32 ymin   = y2clip;
  Sint32 xp, xk, yp, yk, x1, x2, y1, y2, y, plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    while(--plen)
    {
      xk = * f++; yk = * f++;
      if(yp > yk)
      {
        x1 = xk; y1 = yk; x2 = xp; y2 = yp;
      }
      else
      {
        x1 = xp; y1 = yp; x2 = xk; y2 = yk;
      }
      if(((y1 >= y1clip) || (y2 >= y1clip)) && ((y1 <= y2clip) || (y2 <= y2clip)))
      {
        LKElement E;
        if(ymin > y1) ymin = y1;
        if(ymax < y2) ymax = y2;
        E.x  = x1;
        E.dx = (x2 - x1) / (double)(y2 - y1);
        if(y1 < y1clip)
        {
          E.x += E.dx * (y1clip - y1);
          y1 = y1clip;
        }
        E.y = y2;
        if(y1 != y2) LK[y1 - y1clip].push_front(E);
      }
      xp = xk; yp = yk;
    }
  }

  if(ymin < y1clip) ymin = y1clip;
  if(ymax > y2clip) ymax = y2clip;

  for(y = ymin; y <= ymax; y++)
  {
    list<LKElement>::iterator i = LKA.begin();
    while(i != LKA.end()) if(i->y == y) i = LKA.erase(i); else i++;
    if(LK[y - y1clip].size())
    {
      LK[y - y1clip].sort(CompareLKElements);
      LKA.merge(LK[y - y1clip],CompareLKElements);
    }

    bool selector = false;

    for(list<LKElement>::iterator i = LKA.begin(); i != LKA.end(); i++)
    {
      if(selector)
      {
        x2 = (Sint32)floor(i->x);
        gfxClipHLine(s, x1, y, color, x2 - x1 + 1, alpha);
      }
      else x1 = (Sint32)ceil(i->x);
      selector = !selector;
      i->x += i->dx;
    }
    LKA.sort(CompareLKElements);
  }
  LKA.clear();
}

// Rysuje okrg
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// r     - promie okrgu
// color - kolor obrysu
//---------------------------------------

void gfxCircle(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 r, Uint32 color)
{
  Sint32 x = 0, y = r, wx, wy;
  Sint32 e = 0, e1, e2;
  Uint32 w = s->w;
  Uint32 * p = (Uint32 *)(s->pixels) + ys * w + xs;

  wx = x * w;
  wy = y * w;
  while(x <= y)
  {
    * (p + x + wy) = color;
    * (p + y - wx) = color;
    * (p - x - wy) = color;
    * (p - y + wx) = color;
    * (p + y + wx) = color;
    * (p + x - wy) = color;
    * (p - y - wx) = color;
    * (p - x + wy) = color;
    e1 = e  + (x << 1) + 1;
    e2 = e1 - (y << 1) + 1;
    x++; wx += w;
    if(e1 + e2 >= 0)
    {
      e = e2; y--; wy -= w;
    }
    else e = e1;
  }
}

// Ustawia piksel na powierzchni graficznej, uycie wewntrzne
// p - wskanik piksela
// color - kolor piksela
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//---------------------------------------

Uint32 * op;

void gfxPSet(Uint32 * p, Uint32 color, Uint32 alpha)
{
  if(p != op)
  {
    Uint32 c = * p;
    * p = ((((((color & 0xff0000) >> 16) * alpha) + (((c & 0xff0000) >> 16) * (256 - alpha))) & 0x00ff00) << 8) |
          (((((color & 0x00ff00) >>  8) * alpha) + (((c & 0x00ff00) >>  8) * (256 - alpha))) & 0x00ff00) |
          ((((((color & 0x0000ff)) * alpha) + (((c & 0x0000ff)) * (256 - alpha))) & 0x00ff00) >> 8);
    op = p;
  }
}

// Rysuje okrg z przeroczystoci
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// r     - promie okrgu
// color - kolor obrysu
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//---------------------------------------

void gfxCircle(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 r, Uint32 color, Uint32 alpha)
{
  gfxEllipse(s, xs, ys, r, r, color, alpha);  // dla przeroczystoci elipsa jest lepsza
}

// Rysuje okrg, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// r     - promie okrgu
// color - kolor obrysu
//---------------------------------------

void gfxClipCircle(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 r, Uint32 color)
{
  Sint32 x = 0, y = r;
  Sint32 e = 0, e1, e2;

  while(x <= y)
  {
    gfxClipPlot(s, x + xs, y + ys, color);
    gfxClipPlot(s, y + xs,-x + ys, color);
    gfxClipPlot(s,-x + xs,-y + ys, color);
    gfxClipPlot(s,-y + xs, x + ys, color);
    gfxClipPlot(s, y + xs, x + ys, color);
    gfxClipPlot(s, x + xs,-y + ys, color);
    gfxClipPlot(s,-y + xs,-x + ys, color);
    gfxClipPlot(s,-x + xs, y + ys, color);
    e1 = e  + (x << 1) + 1;
    e2 = e1 - (y << 1) + 1;
    x++;
    if(e1 + e2 >= 0)
    {
      e = e2; y--;
    }
    else e = e1;
  }
}

// Rysuje okrg z przeroczystoci, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// r     - promie okrgu
// color - kolor obrysu
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//---------------------------------------

void gfxClipCircle(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 r, Uint32 color, Uint32 alpha)
{
  gfxClipEllipse(s, xs, ys, r, r, color, alpha);  // dla przeroczystoci elipsa jest lepsza
}

// Rysuje pene koo, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// r     - promie okrgu
// color - kolor wypenienia
//---------------------------------------

void gfxFillCircle(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 r, Uint32 color)
{
  Sint32 x = 0, y = r;
  Sint32 e = 0, e1, e2;

  while(x <= y)
  {
    gfxClipHLine(s,-y + xs, x + ys, color, (y << 1) + 1);
    gfxClipHLine(s,-y + xs, -x + ys, color, (y << 1) + 1);
    e1 = e  + (x << 1) + 1;
    e2 = e1 - (y << 1) + 1;
    if(e1 + e2 >= 0)
    {
      gfxClipHLine(s,-x + xs, y + ys, color, (x << 1) + 1);
      gfxClipHLine(s,-x + xs, -y + ys, color, (x << 1) + 1);
      y--; e = e2;
    }
    else e = e1;
    x++;
  }
}

// Rysuje pene koo z przeroczystoci, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// r     - promie okrgu
// color - kolor wypenienia
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//---------------------------------------

void gfxFillCircle(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 r, Uint32 color, Uint32 alpha)
{
  gfxFillEllipse(s, xs, ys, r, r, color, alpha);
}

// Rysuje elips
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// rx    - promie elipsy na osi OX
// ry    - promie elipsy na osi OY
// color - kolor obrysu
//---------------------------------------

void gfxEllipse(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 rx, Sint32 ry, Uint32 color)
{
  Sint32 x = 0, y = ry;
  Sint32 e = 0, e1, e2;
  Sint32 rx2 = rx * rx, ry2 = ry * ry;
  Sint32 fx = 0, fy = rx2 * ry;
  Uint32 w = s->w;
  Uint32 wy;
  Uint32 * p = (Uint32 *)(s->pixels) + ys * w + xs;

  wy = y * w;

  while(fx <= fy)
  {
    * (p + x + wy) = color;
    * (p + x - wy) = color;
    * (p - x + wy) = color;
    * (p - x - wy) = color;
    e1 = e  + (fx << 1) + ry2;
    e2 = e1 - (fy << 1) + rx2;
    x++; fx += ry2;
    if(e1 + e2 >= 0)
    {
      e = e2; y--; fy -= rx2; wy -= w;
    }
    else e = e1;
  }

  while(y >= 0)
  {
    * (p + x + wy) = color;
    * (p + x - wy) = color;
    * (p - x + wy) = color;
    * (p - x - wy) = color;
    e1 = e  - (fy << 1) + rx2;
    e2 = e1 + (fx << 1) + ry2;
    y--; fy -= rx2; wy -= w;
    if(e1 + e2 < 0)
    {
      e = e2; x++; fx += ry2;
    }
    else e = e1;
  }
}

// Rysuje elips z przeroczystoci
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// rx    - promie elipsy na osi OX
// ry    - promie elipsy na osi OY
// color - kolor obrysu
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//---------------------------------------

void gfxEllipse(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 rx, Sint32 ry, Uint32 color, Uint32 alpha)
{
  op = NULL;

  Sint32 x = 0, y = ry;
  Sint32 e = 0, e1, e2;
  Sint32 rx2 = rx * rx, ry2 = ry * ry;
  Sint32 fx = 0, fy = rx2 * ry;
  Uint32 w = s->w;
  Uint32 wy;
  Uint32 * p = (Uint32 *)(s->pixels) + ys * w + xs;

  wy = y * w;

  while(fx <= fy)
  {
    gfxPSet(p + x + wy, color, alpha);
    gfxPSet(p - x + wy, color, alpha);
    gfxPSet(p + x - wy, color, alpha);
    gfxPSet(p - x - wy, color, alpha);
    e1 = e  + (fx << 1) + ry2;
    e2 = e1 - (fy << 1) + rx2;
    x++; fx += ry2;
    if(e1 + e2 >= 0)
    {
      e = e2; y--; fy -= rx2; wy -= w;
    }
    else e = e1;
  }

  while(y >= 0)
  {
    gfxPSet(p + x + wy, color, alpha);
    gfxPSet(p + x - wy, color, alpha);
    gfxPSet(p - x + wy, color, alpha);
    gfxPSet(p - x - wy, color, alpha);
    e1 = e  - (fy << 1) + rx2;
    e2 = e1 + (fx << 1) + ry2;
    y--; fy -= rx2; wy -= w;
    if(e1 + e2 < 0)
    {
      e = e2; x++; fx += ry2;
    }
    else e = e1;
  }
}

// Rysuje elips, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// rx    - promie elipsy na osi OX
// ry    - promie elipsy na osi OY
// color - kolor obrysu
//---------------------------------------

void gfxClipEllipse(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 rx, Sint32 ry, Uint32 color)
{
  Sint32 x = 0, y = ry;
  Sint32 e = 0, e1, e2;
  Sint32 rx2 = rx * rx, ry2 = ry * ry;
  Sint32 fx = 0, fy = rx2 * ry;

  while(fx <= fy)
  {
    gfxClipPlot(s, x + xs, y + ys, color);
    gfxClipPlot(s, x + xs,-y + ys, color);
    if(x) gfxClipPlot(s,-x + xs, y + ys, color);
    if(x) gfxClipPlot(s,-x + xs,-y + ys, color);
    e1 = e  + (fx << 1) + ry2;
    e2 = e1 - (fy << 1) + rx2;
    x++; fx += ry2;
    if(e1 + e2 >= 0)
    {
      e = e2; y--; fy -= rx2;
    }
    else e = e1;
  }

  while(y >= 0)
  {
    gfxClipPlot(s, x + xs, y + ys, color);
    if(y) gfxClipPlot(s, x + xs,-y + ys, color);
    gfxClipPlot(s,-x + xs, y + ys, color);
    if(y) gfxClipPlot(s,-x + xs,-y + ys, color);
    e1 = e  - (fy << 1) + rx2;
    e2 = e1 + (fx << 1) + ry2;
    y--; fy -= rx2;
    if(e1 + e2 < 0)
    {
      e = e2; x++; fx += ry2;
    }
    else e = e1;
  }
}

// Rysuje elips z przeroczystoci, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// rx    - promie elipsy na osi OX
// ry    - promie elipsy na osi OY
// color - kolor obrysu
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//---------------------------------------

void gfxClipEllipse(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 rx, Sint32 ry, Uint32 color, Uint32 alpha)
{
  Sint32 x = 0, y = ry;
  Sint32 e = 0, e1, e2;
  Sint32 rx2 = rx * rx, ry2 = ry * ry;
  Sint32 fx = 0, fy = rx2 * ry;

  while(fx <= fy)
  {
    gfxClipPlot(s, x + xs, y + ys, color, alpha);
    gfxClipPlot(s, x + xs,-y + ys, color, alpha);
    if(x) gfxClipPlot(s,-x + xs, y + ys, color, alpha);
    if(x) gfxClipPlot(s,-x + xs,-y + ys, color, alpha);
    e1 = e  + (fx << 1) + ry2;
    e2 = e1 - (fy << 1) + rx2;
    x++; fx += ry2;
    if(e1 + e2 >= 0)
    {
      e = e2; y--; fy -= rx2;
    }
    else e = e1;
  }

  while(y >= 0)
  {
    gfxClipPlot(s, x + xs, y + ys, color, alpha);
    if(y) gfxClipPlot(s, x + xs,-y + ys, color, alpha);
    gfxClipPlot(s,-x + xs, y + ys, color, alpha);
    if(y) gfxClipPlot(s,-x + xs,-y + ys, color, alpha);
    e1 = e  - (fy << 1) + rx2;
    e2 = e1 + (fx << 1) + ry2;
    y--; fy -= rx2;
    if(e1 + e2 < 0)
    {
      e = e2; x++; fx += ry2;
    }
    else e = e1;
  }
}

// Rysuje wypenion elips, reaguje na prostokat obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// rx    - promie elipsy na osi OX
// ry    - promie elipsy na osi OY
// color - kolor obrysu
//---------------------------------------

void gfxFillEllipse(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 rx, Sint32 ry, Uint32 color)
{
  Sint32 x = 0, y = ry;
  Sint32 e = 0, e1, e2;
  Sint32 rx2 = rx * rx, ry2 = ry * ry;
  Sint32 fx = 0, fy = rx2 * ry;
  while(fx <= fy)
  {
    e1 = e  + (fx << 1) + ry2;
    e2 = e1 - (fy << 1) + rx2;
    if(e1 + e2 >= 0)
    {
      gfxClipHLine(s,-x + xs, y + ys, color,(x << 1) + 1);
      gfxClipHLine(s,-x + xs,-y + ys, color,(x << 1) + 1);
      e = e2; y--; fy -= rx2;
    }
    else e = e1;
    x++; fx += ry2;
  }
  while(y >= 0)
  {
    gfxClipHLine(s,-x + xs, y + ys, color,(x << 1) + 1);
    if(y) gfxClipHLine(s,-x + xs,-y + ys, color,(x << 1) + 1);
    e1 = e  - (fy << 1) + rx2;
    e2 = e1 + (fx << 1) + ry2;
    y--; fy -= rx2;
    if(e1 + e2 < 0)
    {
      e = e2; x++; fx += ry2;
    }
    else e = e1;
  }
}

// Rysuje wypenion elips z przeroczystoci, reaguje na prostokt obcinania
// s     - wskanik struktury SDL_Surface
// xs,ys - wsprzdne rodka
// rx    - promie elipsy na osi OX
// ry    - promie elipsy na osi OY
// color - kolor obrysu
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//---------------------------------------

void gfxFillEllipse(SDL_Surface * s, Sint32 xs, Sint32 ys, Sint32 rx, Sint32 ry, Uint32 color, Uint32 alpha)
{
  Sint32 x = 0, y = ry;
  Sint32 e = 0, e1, e2;
  Sint32 rx2 = rx * rx, ry2 = ry * ry;
  Sint32 fx = 0, fy = rx2 * ry;
  while(fx <= fy)
  {
    e1 = e  + (fx << 1) + ry2;
    e2 = e1 - (fy << 1) + rx2;
    if(e1 + e2 >= 0)
    {
      gfxClipHLine(s,-x + xs, y + ys, color,(x << 1) + 1, alpha);
      gfxClipHLine(s,-x + xs,-y + ys, color,(x << 1) + 1, alpha);
      e = e2; y--; fy -= rx2;
    }
    else e = e1;
    x++; fx += ry2;
  }
  while(y >= 0)
  {
    gfxClipHLine(s,-x + xs, y + ys, color,(x << 1) + 1, alpha);
    if(y) gfxClipHLine(s,-x + xs,-y + ys, color,(x << 1) + 1, alpha);
    e1 = e  - (fy << 1) + rx2;
    e2 = e1 + (fx << 1) + ry2;
    y--; fy -= rx2;
    if(e1 + e2 < 0)
    {
      e = e2; x++; fx += ry2;
    }
    else e = e1;
  }
}

// Rysuje aman
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
//-----------------------------------------

void gfxPoly(SDL_Surface * s, Sint32 * f, Uint32 color)
{
  Sint32 xp,yp,xk,yk,plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    gfxMoveTo(xp,yp);
    while(--plen)
    {
        xk = * f++; yk = *f++;
        gfxLineTo(s,xk,yk,color);
    }
  }
}

// Rysuje aman, reaguje na prostokt obcinania
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
//-----------------------------------------

void gfxClipPoly(SDL_Surface * s, Sint32 * f, Uint32 color)
{
  Sint32 xp,yp,xk,yk,plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    gfxMoveTo(xp,yp);
    while(--plen)
    {
        xk = * f++; yk = *f++;
        gfxClipLineTo(s,xk,yk,color);
    }
  }
}

// Rysuje aman z przeroczystoci
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//-----------------------------------------

void gfxPoly(SDL_Surface * s, Sint32 * f, Uint32 color, Uint32 alpha)
{
  Sint32 xp,yp,xk,yk,plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    gfxMoveTo(xp,yp);
    while(--plen)
    {
        xk = * f++; yk = *f++;
        gfxLineTo(s,xk,yk,color,alpha);
    }
  }
}

// Rysuje aman z przeroczystoci, reaguje na prostokt obcinania
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
// alpha - przeroczysto
//          0   - cakowicie przeroczysty
//          256 - cakowicie nieprzeroczysty
//-----------------------------------------

void gfxClipPoly(SDL_Surface * s, Sint32 * f, Uint32 color, Uint32 alpha)
{
  Sint32 xp,yp,xk,yk,plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    gfxMoveTo(xp,yp);
    while(--plen)
    {
        xk = * f++; yk = *f++;
        gfxClipLineTo(s,xk,yk,color,alpha);
    }
  }
}

// Rysuje lini wygadzan algorytmem Wu Xiaolin
// s     - wskanik struktury SDK_Surface
// x1,y1 - punkt startowy
// x2,y2 - punkt kocowy
// color - kolor linii
//-----------------------------------------

void gfxWuLine(SDL_Surface * s, Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2, Uint32 color)
{
   Uint16 errorpos, error, error2, v;
   Sint32 dx, dy, xdir;

   if (y1 > y2)
   {
      swap(y1,y2);
      swap(x1,x2);
   }

   gfxPlot(s, x1, y1, color);

   if ((dx = x2 - x1) >= 0) xdir = 1;
   else
   {
      xdir = -1;
      dx = -dx;
   }

   if (!(dy = y2 - y1))
   {
      if(xdir == 1)gfxHLine(s, x1, y1, color, dx + 1);
      else         gfxHLine(s, x2, y2, color, dx + 1);
      return;
   }
   if (!dx)
   {
      gfxVLine(s, x1, y1, color, dy + 1);
      return;
   }
   if (abs(dx) == abs(dy))
   {
      gfxLine(s, x1, y1, x2, y2, color);
      return;
   }

   error = 0;

   if(dy > dx)
   {
      errorpos = ((unsigned long) dx << 16) / (unsigned long) dy;
      while (--dy)
      {
         error2 = error;
         error += errorpos;
         if (error <= error2)  x1 += xdir;
         y1++;
         v = (error >> 8) & 0xff;
         gfxPlot(s, x1, y1, color, v ^ 0xff);
         gfxPlot(s, x1 + xdir, y1, color, v);
      }
      gfxClipPlot(s, x2, y2, color);
      return;
   }
   errorpos = ((unsigned long) dy << 16) / (unsigned long) dx;
   while (--dx)
   {
      error2 = error;
      error += errorpos;
      if (error <= error2) y1++;
      x1 += xdir;
      v = (error >> 8) & 0xff;
      gfxPlot(s, x1, y1, color, v ^ 0xff);
      gfxPlot(s, x1, y1 + 1, color, v);
   }
   gfxPlot(s, x2, y2, color);
}

// Rysuje lini wygadzan algorytmem Wu Xiaolin, reaguje na prostokt obcinania
// s     - wskanik struktury SDK_Surface
// x1,y1 - punkt startowy
// x2,y2 - punkt kocowy
// color - kolor linii
//-----------------------------------------

void gfxClipWuLine(SDL_Surface * s, Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2, Uint32 color)
{
   Uint16 errorpos, error, error2, v;
   Sint32 dx, dy, xdir;

   if (y1 > y2)
   {
      swap(y1,y2);
      swap(x1,x2);
   }

   gfxClipPlot(s, x1, y1, color);

   if ((dx = x2 - x1) >= 0) xdir = 1;
   else
   {
      xdir = -1;
      dx = -dx;
   }

   if (!(dy = y2 - y1))
   {
      if(xdir == 1)gfxClipHLine(s, x1, y1, color, dx + 1);
      else         gfxClipHLine(s, x2, y2, color, dx + 1);
      return;
   }
   if (!dx)
   {
      gfxClipVLine(s, x1, y1, color, dy + 1);
      return;
   }
   if (abs(dx) == abs(dy))
   {
      gfxClipLine(s, x1, y1, x2, y2, color);
      return;
   }

   error = 0;

   if(dy > dx)
   {
      errorpos = ((unsigned long) dx << 16) / (unsigned long) dy;
      while (--dy)
      {
         error2 = error;
         error += errorpos;
         if (error <= error2)  x1 += xdir;
         y1++;
         v = (error >> 8) & 0xff;
         gfxClipPlot(s, x1, y1, color, v ^ 0xff);
         gfxClipPlot(s, x1 + xdir, y1, color, v);
      }
      gfxClipPlot(s, x2, y2, color);
      return;
   }
   errorpos = ((unsigned long) dy << 16) / (unsigned long) dx;
   while (--dx)
   {
      error2 = error;
      error += errorpos;
      if (error <= error2) y1++;
      x1 += xdir;
      v = (error >> 8) & 0xff;
      gfxClipPlot(s, x1, y1, color, v ^ 0xff);
      gfxClipPlot(s, x1, y1 + 1, color, v);
   }
   gfxClipPlot(s, x2, y2, color);
}

// Rysuje wygadzony odcinek od zapamitanych wsprzdnych
// do nowych wsprzdnych, ktre po operacji s zapamitywane.
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne koca odcinka
// color  - kod koloru odcinka
//------------------------------------------------------------------

void gfxWuLineTo(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color)
{
   gfxWuLine(screen, gfx_x_coord, gfx_y_coord, x, y, color);
   gfx_x_coord = x; gfx_y_coord = y;
}

// Rysuje wygadzony odcinek od zapamitanych wsprzdnych
// do nowych wsprzdnych, ktre po operacji s zapamitywane. Odcinek
// jest obcinany do prostokta obcinania
// screen - wskanik struktury SDL_Surface
// x,y    - wsprzdne koca odcinka
// color  - kod koloru odcinka
// clip   - wskanik struktury SDL_Rect z prostoktem obcinajcym
//------------------------------------------------------------------

void gfxClipWuLineTo(SDL_Surface * screen, Sint32 x, Sint32 y, Uint32 color)
{
   gfxClipWuLine(screen, gfx_x_coord, gfx_y_coord, x, y, color);
   gfx_x_coord = x; gfx_y_coord = y;
}
// Rysuje wygadzon aman
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
//-----------------------------------------

void gfxWuPoly(SDL_Surface * s, Sint32 * f, Uint32 color)
{
  Sint32 xp,yp,xk,yk,plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    gfxMoveTo(xp,yp);
    while(--plen)
    {
        xk = * f++; yk = *f++;
        gfxWuLineTo(s,xk,yk,color);
    }
  }
}

// Rysuje wygadzon aman, reaguje na prostokt obcinania
// s     - wskanik struktury SDL_Surface
// f     - wskanik listy amanych
// color - kolor wypenienia
//-----------------------------------------

void gfxClipWuPoly(SDL_Surface * s, Sint32 * f, Uint32 color)
{
  Sint32 xp,yp,xk,yk,plen;

  while((plen = * f++))
  {
    xp = * f++; yp = * f++;
    gfxMoveTo(xp,yp);
    while(--plen)
    {
        xk = * f++; yk = *f++;
        gfxClipWuLineTo(s,xk,yk,color);
    }
  }
}
